#include "Elem.h"
#include "Face.h"
#include <cstdio>
#include <alloca.h>
#include "ResizeArray.h"
#include "Connectivity.h"


Connectivity::Connectivity(int _size, int *_pointer, int *_target) {
  size      = _size;
  pointer   = _pointer;
  target    = _target;
  numtarget = pointer[size];
  weight    = 0;
}

Connectivity::Connectivity(ElemSet *els)
{
 int i;
 weight = (float *)0;

 size = els->size();

 // Find out the number of targets we will have
 pointer = new int[size+1] ;
 int pp = 0;
 for(i=0; i < size; ++i) {
   pointer[i] = pp;
   pp += (*els)[i] ? (*els)[i].numNodes() : 0;
 }
 pointer[size] = pp;
 numtarget = pp;

 // Create the target array
 target = new int[pp];

 // Fill it in
 for(i=0; i < size; ++i)
   if((*els)[i])
     (*els)[i].nodes(target+pointer[i]);

}

//HB
Connectivity::Connectivity(FaceSet* fels)
{
 weight = (float *)0;
                                                                                                    
 size = fels->size();
                                                                                                    
 // Find out the number of targets we will have
 pointer = new int[size+1] ;
 int pp = 0;
 for(int i=0; i < size; ++i) {
   pointer[i] = pp;
   pp += (*fels)[i] ? (*fels)[i].numNodes() : 0;
  }
 pointer[size] = pp;
 numtarget = pp;
                                                                                                    
 // Create the target array
 target = new int[pp];
                                                                                                    
 // Fill it in
 for(int i=0; i < size; ++i)
  (*fels)[i].nodes(target+pointer[i]);
}

Connectivity::Connectivity(int _size, int *_count)
{
 size = _size;
 pointer = new int[size+1];
 pointer[0] = 0;
 int i;
 for(i=0; i < _size; ++i)
    pointer[i+1] = pointer[i] + _count[i];
 numtarget = pointer[size];
 target = new int[numtarget];
 weight = NULL;
}

Connectivity::~Connectivity()
{
 delete [] pointer;
 delete [] target;
 if(weight) delete [] weight;
}

// reverse() return a new connectivity that is the reverse of the present one
Connectivity *
Connectivity::reverse(float *w)  
{
 Connectivity *res = new Connectivity();
 // The reverse connecivity has the same size as the original
 res->numtarget = numtarget;
 res->target = new int[numtarget];
 res->weight = w;

 // Find the max of target
 int maxtarg = 0;
 int i;
 for(i=0; i < numtarget; ++i) {
   if(target[i] > maxtarg) maxtarg = target[i];
   if(target[i] < 0) fprintf(stderr, "Inconsistent data found\n");
 }

 res->size = maxtarg+1;
 res->pointer = new int[res->size+1];
 for(i = 0; i <= res->size; ++i)
   res->pointer[i] = 0;

 // Now do a first pass to fill in res->pointer
  for(i=0; i < numtarget; ++i)
    res->pointer[target[i]]++;

 for(i = 1; i <= res->size; ++i)
   res->pointer[i] += res->pointer[i-1];

 // Second pass fills in target
 int j;
 for(i=0; i < size; ++i)
  {
   int jend = pointer[i+1];
   for(j = pointer[i]; j < jend; ++j)
    {
     int tg = target[j];
    res->target[--res->pointer[tg]] = i;
    }
  }

 return res;
}

int
Connectivity::offset(int i, int j)
{
 int ii;
 for(ii = pointer[i]; ii < pointer[i+1];++ii)
  if(target[ii] == j) return ii;

 return -1; // We didn't find a connection between i and j
}

Connectivity* 
Connectivity::transcon(Connectivity* tc) 
{
 int i,j,k;

 // First find the biggest target so we can size arrays correctly
 int tgmax=-1;

 for(i =0; i < tc->numtarget; ++i)
   if(tc->target[i] > tgmax) tgmax=tc->target[i];
 tgmax++; // Important adjustment

 // Now we can size the array that flags if a target has been visited
 int *flags = new int[tgmax];
 for(i = 0; i < tgmax; ++i)
   flags[i] = -1;

 // Compute the new pointers
 int *np = new int[size+1];
 int cp = 0;
 for(i = 0; i < size; ++i) {
  np[i] = cp;
  for(j = pointer[i]; j < pointer[i+1]; ++j)
    {
      int intermed = target[j];
      for (k = 0; k < tc->num(intermed); ++k)
        if(flags[(*tc)[intermed][k]] != i) {
          flags[(*tc)[intermed][k]] = i;
          cp ++;
        }
    }
 }
 np[size] = cp;
 // Now allocate and fill the new target
 for(i = 0; i < tgmax; ++i)
   flags[i] = -1;
 int *ntg = new int[cp];

 cp = 0;

 for(i = 0; i < size; ++i) {
  for(j = pointer[i]; j < pointer[i+1]; ++j)
    {
      int intermed = target[j];
      for (k = 0; k < tc->num(intermed); ++k)
        if(flags[(*tc)[intermed][k]] != i) {
          flags[(*tc)[intermed][k]] = i;
          ntg[cp] = (*tc)[intermed][k];
          cp ++;
        }
    }
 }

 delete[]flags;

 Connectivity *res = new Connectivity();
 res->size      = size;
 res->pointer   = np;
 res->numtarget = np[size];
 res->target    = ntg;
 res->weight    = 0;
 return res;
}


int
Connectivity::num(int nd, int *mask)
{
 int res=0;
 int jstrt = pointer[nd];
 int jstop = pointer[nd+1];
 int j;
 for(j = jstrt; j < jstop; ++j)
    if(mask[target[j]]) res++;
 return res;
}

void
Connectivity::print(FILE *f)
{
 int i, fortran = 0;
 int maxdist = 0;
 fprintf(f, "Connectivity size %d\n",size);
 fflush(f);
 for(i = 0; i < size; ++i) {
   fprintf(f, "%d ->\n", i+fortran);
   int j;
   for(j = pointer[i]; j < pointer[i+1]; ++j) {
     fprintf(f, " %d\n", target[j]+fortran);
     if(i-target[j] > maxdist ) maxdist = i-target[j];
   }
   fprintf(f,"\n");
 }
 fprintf(f, "Max dist %d\n", maxdist);
 fflush(f);
}

void
Connectivity::print(char *fn)
{
 FILE *f=fopen(fn,"w");

 int i, fortran = 0;
 int maxdist = 0;
 fprintf(f, "Connectivity size %d\n",size);
 fflush(f);
 for(i = 0; i < size; ++i) {
   fprintf(f, "%d ->\n", i+fortran);
   int j;
   for(j = pointer[i]; j < pointer[i+1]; ++j) {
     fprintf(f, " %d\n", target[j]+fortran);
     if(i-target[j] > maxdist ) maxdist = i-target[j];
   }
   fprintf(f,"\n");
 }
 fprintf(f, "Max dist %d\n", maxdist);
 fflush(f);

 fclose(f);
}

void
Connectivity::findPseudoDiam(int *s, int *e, int *mask)
{
 int i,k,nw;
 // Select the node with the lowest connectivity
 int cmin = numtarget+1;
 int cmax = 0;
 for(i = 0; i < size; ++i) {
  if(mask[i]) {
     if((nw = num(i,mask)) < cmin) {
       cmin=nw;
       *s = i;
     }
    if(nw > cmax) cmax = nw;
  }
 }

 int *ls = (int *) alloca(numtarget*sizeof(int));
 int *xls = (int *) alloca((numtarget+1)*sizeof(int));
 // Created rooted level structure
 int w;
 int h = rootLS(*s, xls, ls, w, mask);
 int subconsize = xls[h];
 int *sorted = (int *) alloca((cmax+1)*sizeof(int));
 *e = ls[xls[h-1]]; // Give a default end point in case h == subconsize.
 while(h < subconsize) {
   for(k=0; k <= cmax; ++k) sorted[k] = -1;
   // Find in the last level the node with the minimum connectivity
   int maxweight = subconsize;
   int kstrt = xls[h-1];
   int kstop = xls[h];
   for(k=kstrt; k < kstop; ++k)
      {
        sorted[num(ls[k],mask)] = ls[k];
      }
   int w_e = subconsize;
   for(k = 0; k <= cmax; ++k)
    if(sorted[k] >= 0) {
      int nh = rootLS(sorted[k], xls, ls, w, mask);
      if(w < w_e) {
         if(nh > h) {
           *s = sorted[k];
           h = nh;
           break;
         }
         *e = sorted[k];
         w_e = w;
      }
    }
   if(k > cmax) break;
 }
 return;
}

int
Connectivity::rootLS(int root, int *xls, int *ls, int &w, int *mask)
{
 int i, j;
 w = 0;
 int *locMask = (int *) alloca(size*sizeof(int));
 if(mask)
   for(i = 0; i < size; ++i) locMask[i] = mask[i];
 else
   for(i = 0; i < size; ++i) locMask[i] = 1;

 locMask[root] = 0;
 ls[0] = root;
 xls[0] = 0;
 int nlvl = 1;
 int nf = 1;
 while(nf > xls[nlvl-1]) {
   xls[nlvl] = nf;
   int lbegin = xls[nlvl-1];
   int lend   = xls[nlvl];
   for (i = lbegin; i <lend; ++i) {
     int n1 = ls[i];
     int jstart = pointer[n1];
     int jstop  = pointer[n1+1];
     for(j=jstart; j < jstop; ++j) {
       int n2 = target[j];
       if(locMask[n2]) {
          locMask[n2] = 0;
          ls[nf++] = n2;
       }
     }
   }
   if(nf-xls[nlvl] > w) w = nf-xls[nlvl];
   nlvl++;
 }
 return nlvl-1;
}

compStruct
Connectivity::renumByComponent(int renumAlg)
{

// size = total number of nodes

  int *globalRenum = new int[size];
  int *mark        = (int *) alloca(size*sizeof(int));

  int *ls         = (int *) alloca(size*sizeof(int));
  int *xls        = (int *) alloca((size+1)*sizeof(int));


  ResizeArray<int> xcomp(0,2);
// Initialize mark to zero, accounting for missing node #s
// Initialize globalMask

  int inode;
  for(inode = 0; inode < size; ++inode) {
    mark[inode] = (num(inode) != 0) ? 1 : 0;
    globalRenum[inode] = -1;
  }

// Loop over nodes checking which ones are marked
// and belong to the same component.

  int j, k, nextNum = 0, count = 0;
  int *locMask = (int *) alloca(size*sizeof(int));
  for(inode = 0; inode < size; ++inode)
     locMask[inode] = 0;

  int currentNumber = 0;
  xcomp[0] = currentNumber;

  for(inode = 0; inode < size; ++inode)
  {
    if(mark[inode] == 1)
    {
      // Find all neighbors of inode
 
      int w;
      int h = rootLS(inode, xls, ls, w);

      // Declare and set local mask

      for(j=0; j<xls[h]; ++j) {
          locMask[ls[j]] = 1;
          mark[ls[j]] = 0;
      }

      // call renumbering for local mask
      int *lrenum = (int *) alloca(size*sizeof(int));
      if(renumAlg) renumSloan(locMask, nextNum, lrenum);

      // reset locMask to zero
      for(j=0; j<xls[h]; ++j)
        locMask[ls[j]] = 0;

      // Assemble local mask into global mask
      if(renumAlg) { 
        for(j=0; j<xls[h]; ++j) {
            k = ls[j];
            globalRenum[k] = lrenum[k];
        }
      }
      else {
        for(j=0; j<xls[h]; ++j) {
           k = ls[j];
           globalRenum[k] = currentNumber + j;
        }
      }
      currentNumber += xls[h];
      count += 1;
      xcomp[count] = currentNumber;
    }
  }

  fprintf(stderr," ... Number of components =%2d       ...\n",count);

  compStruct ret;
  ret.numComp = count;
  ret.xcomp = xcomp.yield(); 
  ret.renum = globalRenum;

  return ret;
}


int *
Connectivity::renumSloan(int *mask, int &nextNum, int *renum)
{
 int i,j,k;
 int s_node, e_node;
 float w1=1.0, w2=2.0;

 if(mask == 0)
  {
    mask = (int *) alloca(sizeof(int)*size);
    for(i=0; i < size; ++i)
       mask[i] = (num(i)) ? 1 : 0;
  }

 findPseudoDiam( &s_node, &e_node, mask);

 int *ls  = (int *) alloca(sizeof(int)*numtarget);
 int *xls = (int *) alloca(sizeof(int)*(numtarget+1));

 int w;
 int h = rootLS(e_node, xls,ls,w,mask);
 // now give a distance to each point

 if(renum == NULL) renum = new int[size];
 int *distance = renum;
 int *status = distance;

 for(i = 0; i < size; ++i)
   distance[i] = -1;

 for(i = 0; i < h; ++i) {
    for(j = xls[i]; j < xls[i+1]; ++j)
      distance[ls[j]] = i;
  }

 float *priority = (float *) alloca(sizeof(float)*size);
 // initialize the priority values
 for(i = 0; i < size; ++i) {
   if(mask[i]) {
     priority[i] = w1*distance[i] - w2*num(i);
     distance[i] = -2; // status and distance are stored in the same array
   }
 }

 // initalize the queue with the starting point
 int *q = (int *) alloca(sizeof(int)*size);
       // maximum size is all the points in the q
 q[0] = s_node;
 int nn=1;
 status[s_node] = -1;

 // While the queue is not empty
 while(nn > 0) {
   // Find in the queue the point with maximum priority
   int next = 0;
   float maxpri = priority[q[next]];
   for(i = 1; i < nn; ++i) {
      if(priority[q[i]] > maxpri) {
         maxpri = priority[q[i]];
         next = i;
      }
   }
   // Remove the next node numbered from the queue
   int nextnode = q[next];
   q[next] = q[nn-1];
   nn = nn-1;
   int istart = pointer[nextnode], istop = pointer[nextnode+1];
   if(status[nextnode] == -1) {
     status[nextnode] = 0; // So we don't step on our feet.
     // Preactive node. Examine its neighbors
     for(i = istart; i < istop; ++i) {
        int neighbor = target[i];
        priority[neighbor] += w2; // update the priority
        if(status[neighbor] == -2) { // if neighbor was inactive, it becomes pre-active
           // NB the next loop will set the same nodes as active
           q[nn] = neighbor;
           nn++;
           status[neighbor] = -1;
        }
     }
   }
   status[nextnode] = nextNum;
   nextNum++;
   // Now scan the preactive neighbors, make them active and their neighbors become
   // preactive if they were inactive
   for(i = istart; i < istop; ++i) {
      int neighbor = target[i];
      if(status[neighbor] == -1) {
        status[neighbor] = 0;
        priority[neighbor] += w2;
        int kstart = pointer[neighbor], kstop = pointer[neighbor+1];
        for(k = kstart; k < kstop ; ++k) {
            int kn = target[k];
            priority[kn] += w2;
            if(status[kn] == -2) { // This node is inactive. must become preactive
               status[kn] = -1;
               q[nn] = kn;
               nn++;
            }
        }
      }
    }
 }
 return status;
}
#ifdef _RCM_
int *
Connectivity::gen_renum(int numnod)
{
  Connectivity *ne = reverse();
  Connectivity *nn = ne->transcon(this);
  // int numnod = nn->size();
  int * perm = new int[numnod+2];
  int * inv_perm = new int[numnod+2];
  int * xls = new int[numnod+2];
  int * renum = new int[numnod];

  int *mask = new int[numnod];
  int i;
  for(i=0; i < numnod; ++i)
    mask[i] = (ee->num(i) == 0) ? 0 : 1;
  genrcm(numnod,ee->pointer,ee->target,perm,mask,xls,inv_perm);

  for (i=1; i<=numnod; i++)
    if (renum[i] != -1)
      renum[i] = inv_perm[i];
  delete ne;
  delete nn;
  delete perm;
  delete inv_perm;
  delete xls;
}

void degree(int root,int xadj[],int adjncy[],int mask[],
int deg[],int *ccsize,int ls[])
{
  int i,ideg,j,jstop,jstrt,lbegin,lvlend,lvsize,nbr,node;

  ls[1] = root;
  xadj[root] = -xadj[root];
  lvlend = 0;
  *ccsize = 1;

  a100: lbegin = lvlend + 1;
        lvlend = *ccsize;

        for( i=lbegin; i<=lvlend; i++)
           {
             node = ls[i];
             jstrt = -xadj[node];
             jstop = abs(xadj[node+1])-1;
             ideg = 0;
             if(jstop<jstrt) goto a300;

             for( j = jstrt; j<=jstop; j++)
                {
                  nbr = adjncy[j];
                  if(mask[nbr] == 0) goto a200;
                  ideg = ideg + 1;
                  if(xadj[nbr]<0) goto a200;
                  xadj[nbr] = -xadj[nbr];
                  *ccsize = *ccsize+1;
                  ls[*ccsize] = nbr;
                  a200:   ;
                }

              a300:   deg[node] = ideg;
            }
         lvsize = *ccsize - lvlend;

         if(lvsize>0) goto a100;

         for(i=1;i<=*ccsize;i++)
         {
           node = ls[i];
           xadj[node] = - xadj[node];
         }
}

/* ------------------------------------------------------------------- */

void rcm(int root,int *xadj,int *adjncy,int *mask,
int *perm,int *ccsize,int *deg ,int *inv_perm,int ishift)

{
   int fnbr,i,j,jstop,jstrt,k,l,lbegin,lnbr,lperm,lvlend,nbr,node;

   degree(root,xadj,adjncy,mask,deg,ccsize,perm);

   mask[root] = 0;
   lvlend = 0;
   lnbr = 1;

a100:   lbegin = lvlend + 1;
        lvlend = lnbr;
        for( i = lbegin; i<= lvlend; i++ )
           {
             node = perm[i];
             jstrt = xadj[node];
             jstop = xadj[node+1] - 1;
             fnbr = lnbr + 1;
             for( j=jstrt; j<=jstop; j++)
                {
                  nbr = adjncy[j];
                  if(mask[nbr]==0) goto a200;
                  lnbr = lnbr + 1;
                  mask[nbr] = 0;
                  perm[lnbr] = nbr;
                  inv_perm[nbr] = lnbr + ishift;

                  a200: ;
                }
             if(fnbr>=lnbr) goto a600;
             k = fnbr;

             a300 :  l = k;
                     k = k+1;
                     nbr=perm[k];

             a400:   if(l<fnbr) goto a500;
                     lperm = perm[l];
                     if(deg[lperm]<=deg[nbr]) goto a500;
                     perm[l+1] = lperm;
                     inv_perm[lperm] = l + 1 + ishift;
                     l = l-1;
                     goto a400;

             a500:   perm[l+1] = nbr;
                     inv_perm[nbr] = l + 1 + ishift;
                    if(k<lnbr) goto a300;

             a600:   ;
         }

      if(lnbr>lvlend) goto a100;
      k=(*ccsize)/2;
      l = (*ccsize);
      for( i=1; i<= k; i++)
         {
           lperm = perm[l];
           perm[l] = perm[i];
           inv_perm[perm[i]] = l + ishift;
           perm[i] = lperm;
           inv_perm[lperm] = i + ishift;
           l = l-1;
         }
}

/* ------------------------------------------------------------------- */


void rootls(int root,int xadj[ ],int adjncy[ ],int mask[ ],int *nlvl,
int xls[ ],int ls[ ])
{
int i,j,jstop,jstrt,lbegin,ccsize,lvlend,lvsize,nbr,node;

mask[root] = 0;
ls[1] = root;
*nlvl = 0;
lvlend = 0;
ccsize = 1;

a_200:
lbegin = lvlend+1;
lvlend = ccsize;
*nlvl = *nlvl + 1;
xls[*nlvl]= lbegin;

for (i = lbegin; i <= lvlend; i++)
   {
     node = ls[i];
     jstrt= xadj[node];
     jstop= xadj[node+1] - 1;
     if(jstop < jstrt) continue;
     for (j = jstrt; j <= jstop; j++)
        {
          nbr = adjncy[j];
          if(mask[nbr] == 0) continue;
          ccsize = ccsize+1;
          ls[ccsize] = nbr;
          mask[nbr] = 0;
        }
   }

lvsize = ccsize - lvlend;
if(lvsize > 0) goto a_200;
xls[*nlvl+1] = lvlend+1;

for (i = 1; i <= ccsize; i++)
   {
     node = ls[i];
     mask[node] = 1;
   }

}


/* ------------------------------------------------------------------- */


void fnroot(int *root,int xadj[ ],int adjncy[ ],int mask[ ],
int nlvl,int xls[ ],int ls[ ])
{
int ccsize,j,jstrt,k,kstop,kstrt,mindeg,nabor,ndeg,node,nunlvl;

rootls(*root,xadj,adjncy,mask,&nlvl,xls,ls);
ccsize = xls[nlvl+1] - 1;

a_100:
jstrt = xls[nlvl];
mindeg = ccsize;
*root = ls[jstrt];
if(ccsize == jstrt) goto a_400;

for (j = jstrt; j <= ccsize; j++)
   {
     node = ls[j];
     ndeg = 0;
     kstrt = xadj[node];
     kstop = xadj[node+1] - 1;
     for (k = kstrt; k <= kstop; k++)
        {
          nabor = adjncy[k];
          if(mask[nabor] > 0) ndeg = ndeg+1;
        }
     if(ndeg >= mindeg) continue;
     *root = node;
     mindeg= ndeg;
   }
a_400:
rootls(*root,xadj,adjncy,mask,&nunlvl,xls,ls);
if(nunlvl <= nlvl) goto a_900;
nlvl = nunlvl;
if(nlvl < ccsize) goto a_100;

a_900:  ;

}

/* ------------------------------------------------- */

void genrcm ( int neqns,int xadj[],int adjncy[],int perm[],int mask[],int xls[],int inv_perm[])
{
   int ccsize, i, nlvl = 0, num, root, ishift;

/* one change from text book:  must set mask before calling genrcm.
   Used for masking out interface nodes                            */

   num = 1; ishift = 0;
   for (i=0; i<neqns; i++)
   {
      if (mask[i] == 0)
         goto a_201;
      else
      {
         root = i;
         fnroot(&root, xadj, adjncy, mask, nlvl, xls, &perm[num-1]);
         rcm(root, xadj, adjncy, mask, &perm[num-1], &ccsize, xls, inv_perm,ishift);
         num += ccsize; ishift += ccsize;
         if (num > neqns) goto a_901 ;
      }
      a_201: ;
   }
   a_901:   ;
}
#endif
