#ifdef _OPENMP
#include <omp.h>
#endif

#include"Point3D.h"
#include"Face.h"
#include"Elem.h"
#include"Decomp.h"
/*
#include<iostream>
using std::cerr;
using std::endl;
#include<iomanip>
#include<fstream>
using std::ifstream;
using std::ofstream;
using std::ios;
*/
#include<cstdio>
#include<cstdlib>
#include<cstring>
#include"aux.h"
#include"Connectivity.h"
#include"Decomp3D.h"
#include"Decomp2D.h"
#include"Timer.h"
#include "values.h"

//external variable of an abstracted type containing the structural dom's
//information, declared in aux.C
extern Domain* domain;
//file ptr needed by the parser, declared in aux.C
extern FILE* yyin;

extern int readCommandLine(int, char **, FileNames &, int &, int &, int &, int &, double &, double &, int &);

//--------------------------------------------------------------------

//global variable that describes the dimension of the interface
int dimension = 2;
bool beamFlag = false;

int main(int argc, char* argv[])  {

  //system("clear");
  Timer progTimer;
  double progTime = progTimer.getTime();

  //iterators
  int i,j,k;

  //flag used to signal CNBC has been set in an input file
  int bcFlag=0, decomposeFlag=0, noOverlap=0, numThreads = 1; 
  //HB: change to 1 so that if not specified (see -p otption on command line),
  //    an OpenMP executable behaves as a serial executatble
//#ifdef _OPENMP
//  numThreads = 2; 
//#endif

  //temp vars: fluid domain #, struct domain #, node #
  int fnum,nnum,snum;

  //number of struct/fluid subdomains
  int numStructSub, numFluidSub;

  //tolerance for the matching routines, can be changed with -t flag
  double tol = 0.1;
  double maxdist = MAXDOUBLE;
  int numtol = 1;

  //buffer used for input when parsing files
  char buffer[80];

  char outFile[9], outname[500];

  FileNames fn;

  // read command line arguments
  int type = readCommandLine(argc, argv, fn, decomposeFlag, bcFlag, 
			     noOverlap, numThreads, tol, maxdist, numtol);
#ifdef _OPENMP
  fprintf(stderr,"... Uses %2d threads\n",numThreads);
#endif
  double *tols = new double[numtol];
  double *maxdists = new double[numtol];
  for(int i=0; i<numtol; ++i) {
    tols[i] = tol*double(i+1);
    maxdists[i] = maxdist*double(i+1);
  }

  // create decomp class
  Decomp* decomp;  //Decomp is actually a virtual class that we abstract into

  //two different types: 2D and 3D
  if(dimension == 2)
    decomp = new Decomp3D;
  else 
    decomp = new Decomp2D;
 
  // read the fluid mesh into memory
  fprintf(stderr, "... Reading fluid geometry input file .............. %s\n", fn.fluid);
  Timer miscTimer;
  double miscTime = miscTimer.getTime();
  decomp->getFluidDomain(fn.fluid);
  miscTime = miscTimer.getTime() - miscTime;
  fprintf(stderr, "... Fluid geometry input file read time ............ %f s.\n", miscTime);

  // check if only decomposition is desired
  if(decomposeFlag == 1)  {
    fprintf(stderr, "... Only performing fluid decomposition\n");
   
    // read fluid decomposition file into memory
    decomp->getDecomp(fn.decF);
    if(fn.decF!=NULL)
      fprintf(stderr, "... Reading fluid decomposition file ............ %s\n", fn.decF);
    else
      fprintf(stderr, "... No Fluid Decomposition file: Assuming 1 Fluid Subdomain\n");

    // Form all the subdomains according to decomp file
    decomp->makeSubD(noOverlap);
    fprintf(stderr, "... Fluid subdomains constructed\n");

    // Find the interfaces for each subdomain
    decomp->getDomainConnect();
    if (type == 0)
      fprintf(stderr, "... Fluid subdomains' interfaces found\n");

    // output the flu- and flu.glob files
    decomp->print();
    fprintf(stderr, "... Outputted the flu- and flu.glob files\n");
    fprintf(stderr, "************DONE*********************\n");
    return 0;
  }
  
  // Attempt to open Structure Mesh File
  FILE* structFilePtr;
  structFilePtr=fopen(fn.structure,"r");
  if(!structFilePtr) {
    fprintf(stderr, "*** ERROR: Can't Open Structure Input File: %s\n", fn.structure);
    fprintf(stderr, "... Exiting\n");
    exit(1);
  }

  // Parses structMesh using lex and yacc.
  // This function is a member of the Domain class and
  // can be found in aux.{h,C} the lex and yacc scripts can be found
  // in p.y and leyer.l
  yyin = structFilePtr;

  fprintf(stderr, "... Reading structure input file ................... %s\n", fn.structure);
  miscTime = miscTimer.getTime();
  buildDomain();
  miscTime = miscTimer.getTime() - miscTime;
  fprintf(stderr, "... Structure input file read time ................. %f s.\n", miscTime);


  // read struct control file
  //the remaing auxFile info is read using readAux. 
  if(fn.control!=NULL){
    domain->readAux(fn.control);
    fprintf(stderr, "... Reading structure control file: %s\n", fn.control);
  }

  // Read struct decomposition file
  domain->getDecomp(fn.decS);
 
  // Return the number of structural subdomains
  numStructSub=domain->getNumSub();	

  if(fn.decS != NULL)
    fprintf(stderr,"... Reading structural decomposition from: %s\n", fn.decS);

  // A struct containing the fluid mesh info is returned to main
  FluidDomain* fluidDom = decomp->getDomain();

  // nodeBC will contain the node bound. cond.
  // in the 2D case, these are provided directly by the user
  // in 3D, they must be abstracted from the BC of the faces
  int* nodeBC = new int [fluidDom->nodes.size()];
  int warnFlg = 0;

  // 3D case
  miscTime = miscTimer.getTime();;
  if (dimension == 2){	

    for (i = 0; i < fluidDom->nodes.size(); i++)
      nodeBC[i] = 0;

    for (i=0; i<fluidDom->faces.size(); i++) {
      Face &face = fluidDom->faces[i];

      for (k=0; k<face.numNodes(); k++) {

	if (nodeBC[ face[k] ] > 0 && fluidDom->getTypeBC(i) < 0)
	  warnFlg = 1;

	if ( (nodeBC[ face[k] ] == 0) ||
	     (nodeBC[ face[k] ] > 0 && fluidDom->getTypeBC(i) < 0))
	  nodeBC[ face[k] ] =  fluidDom->getTypeBC(i);	
      }
    }

    if (warnFlg) 
      fprintf(stderr, "*** Warning: moving boundary conditions have priority\n");

  } else {

    for (i = 0; i < fluidDom->nodes.size(); i++)
      nodeBC[i] = fluidDom->getTypeBC(i);

  }

  // now read any user specified list of node bc's to be changed
  if (bcFlag == 1)  {
    ifstream bcIn(fn.bcFile,ios::in);
    bcIn >> buffer;
    if (strncmp(buffer,"CNBC",4) == 0)  {
      int numChange, flag=1, nodeNum;
      bcIn >> numChange;

      //numChange refers to a user supplied number whose modulus
      //is the number of nodes whose BC will be changed
      //if numChange is positive these nodes will be turned "on"
      //if negative they will be turned "off"
      if(numChange>0)
	flag=-1;
      numChange = abs(numChange);
      for(i=0; i<numChange; i++)  {
	bcIn>>nodeNum;
	nodeBC[nodeNum-1]=flag;
      }
    }
    else  {
      fprintf(stderr, "*** ERROR: Can't read file: %s\n", fn.bcFile);
      fprintf(stderr, "... Exiting\n");
      exit(1);
    }
  }

  //fprintf(stderr, "...Boundary conditions determined\n");

  // PJSA: open and read matched elements input file
  int numMatchedNEPairs = 0;
  ResizeArray<int> matchedFluidNodesE(0), matchedStructureElements(0);
  if (fn.matchedElements != NULL) {
    ifstream matchedElementsFile(fn.matchedElements,ios::in);

    if(!matchedElementsFile) {
      cerr << "*** ERROR: Can't Open Matched Elements Input File: " << fn.matchedElements << endl;
      cerr << "... Exiting\n";
      exit(1);
    }

    cerr << " ... Reading Matched Elements Input File ... \n";

    while( (matchedElementsFile >> matchedFluidNodesE[numMatchedNEPairs]) && 
           (matchedElementsFile >> matchedStructureElements[numMatchedNEPairs]) ) 
      numMatchedNEPairs++;

    cerr << " ... Read " << numMatchedNEPairs << " matching node-element pairs \n";

    matchedElementsFile.close();
  }

  // PJSA: open and read matched nodes input file
  int numMatchedNNPairs = 0;
  ResizeArray<int> matchedFluidNodesN(0), matchedStructureNodes(0);
  if (fn.matchedNodes != NULL) {
    ifstream matchedNodesFile(fn.matchedNodes,ios::in);

    if(!matchedNodesFile) {
      cerr << "*** ERROR: Can't Open Matched Nodes Input File: " << fn.matchedNodes << endl;
      cerr << "... Exiting\n";
      exit(1);
    }

    cerr << " ... Reading Matched Nodes Input File ... \n";

    while( (matchedNodesFile >> matchedFluidNodesN[numMatchedNNPairs]) &&
           (matchedNodesFile >> matchedStructureNodes[numMatchedNNPairs]) ) 
      numMatchedNNPairs++;

    cerr << " ... Read " << numMatchedNNPairs << " matching node-node pairs \n";

    matchedNodesFile.close();
  }

  // Now the actual matching is done and a TOP/DOMDEC
  // format file called match.top is output
  // to be matched a node must have a negative BC
  // and the match must satisy criterion set by the	
  // applicable matching routine	
  // matchPoint can be found in aux.{h,C} but the actual
  // matching is done in the classes abstracted from 
  // ElementClass  
  // match.top file simply gives a description of the 
  // lines connecting each match

  // struct containing information about a match (success,struct subD of
  // matched elem, natural coords of match, elem number, fluid node num)
  interCoords *coords = new interCoords[fluidDom->nodes.size()];

  // filter out interface nodes
  int *interfNode = new int[fluidDom->nodes.size()];
  int count = 0;
  for (i =0; i<fluidDom->nodes.size(); i++) {
    coords[i].success = 0;

    if(nodeBC[i]<0)
      interfNode[count++] = i;
  }

  miscTime = miscTimer.getTime() - miscTime;
  fprintf(stderr, "... Interface extraction completion time ........... %f s.\n", miscTime);
  
  // for timing
  Timer parallelTimer;
  
#ifdef _OPENMP
  omp_set_num_threads(numThreads);
#endif
  
#define MAXLINE 500
  char format[MAXLINE];
  strcpy(format, "\b");
  for (k=1; k<MAXLINE-1; ++k)
    strcat(format, "\b");

  int numMatched = 0;
  double totalParallelTime = 0.0;
  for(j=0; j<numtol; ++j) {
    int pc_step = 1 * (count/numThreads) / 100;
    int thread_num = 0;
    int thread_count = 0;
    double parallelTime = parallelTimer.getTime();

    fprintf(stderr, "... Pass %d\n",j);

    if(maxdists[j] != MAXDOUBLE) 
      fprintf(stderr, "... Matching %d interface fluid nodes (%d threads, tol=%g, maxdist=%f)\n",
	      count-numMatched, numThreads, tols[j], maxdists[j]);
    else 
      fprintf(stderr, "... Matching %d interface fluid nodes (%d threads, tol=%g, unlimited distance)\n",
	      count-numMatched, numThreads, tols[j]);

#pragma omp parallel for shared(thread_count)
    for (i=0; i<count; i++)  {
      if (!coords[ interfNode[i] ].success) {
	domain->matchPoint(fluidDom->nodes[interfNode[i]], tols[j],
			   coords[interfNode[i]], maxdists[j]);

	if (coords[ interfNode[i] ].success) 
	  numMatched++;
      }

#ifdef _OPENMP
      thread_num = omp_get_thread_num();
#endif

      if (thread_num == 0) {
	if (pc_step && thread_count % pc_step == 0) {
	  fprintf(stdout, "    Processed %d%% of surface nodes", 100*thread_count/(count/numThreads));
	  fprintf(stdout, format);
	  fflush(stdout);
	}
	++thread_count;
      }
    }   
  
    // HB: I think the searches below are done "numtol" times but with a fixed tolerance 
    // PJSA: process matched nodes read from input file
    // this will override any previous match found
    for (i=0; i<numMatchedNEPairs; ++i) {
      domain->matchPoint(matchedStructureElements[i], fluidDom->nodes[matchedFluidNodesE[i]],
			 tol, coords[matchedFluidNodesE[i]]);
    }
    
    // PJSA: process matched nodes read from input file
    // this will override any previous match found
    for (i=0; i<numMatchedNNPairs; ++i) {
      domain->matchPoint(matchedStructureNodes[i], coords[matchedFluidNodesN[i]]);
    }
    
    parallelTime = parallelTimer.getTime() - parallelTime;
    totalParallelTime += parallelTime;
    fprintf(stderr,"\n... Matching completion time ....................... %f s.\n", parallelTime);
    if (numMatched >= count) 
      break;
  }
  delete [] tols;
  delete [] maxdists;  
	
  sprintf(outname, "%s.match.top", fn.outFile);
  FILE* matchOut = fopen(outname, "w");
  fprintf(matchOut, "Nodes MatchNodes\n");
  fprintf(stderr,"... Outputting xpost matcher file .................. %s\n", outname);
  int numOut=0;
  int numNonMatched = 0; //HB: count number of non-matched fluid nodes
  for (i=0; i<count; i++) {
    if (coords[ interfNode[i] ].success == 1) {
      fprintf(matchOut, "%8d %g %g %g\n", ++numOut, fluidDom->nodes[ interfNode[i] ][0],
	      fluidDom->nodes[ interfNode[i] ][1], fluidDom->nodes[ interfNode[i] ][2]);
      fprintf(matchOut, "%8d %g %g %g\n", ++numOut, coords[ interfNode[i] ].point[0],
	      coords[ interfNode[i] ].point[1], coords[ interfNode[i] ].point[2]);
    }
    else {
      numNonMatched++;
      //cerr<<"*** WARNING: Node "<<interfNode[i]+1<<" not matched!"<<endl;
    }
  }
  
  fprintf(matchOut, "Elements MatchElements using MatchNodes\n");
  numOut=numOut/2;
  for (i=1;i<=numOut;i++)
    fprintf(matchOut, "%8d 1 %8d %8d\n", i, 2*i-1, 2*i);
  fclose(matchOut);
  
  // HB: write to file the Id of the non-matched fluid nodes
  // To visualize them, it may be usefull to also include them in the preceeding or new
  // xpost file by adding an ElementSet (for instance called NonMachedNodes ...) where 
  // the 2 nodes bar els have identical nodes

  int nonMatchNum = 0;
  if(numNonMatched) {
    fprintf(stderr,"*** WARNING: %6d nodes have not been matched.\n",numNonMatched);
    sprintf(outname, "%s.bcdata", fn.outFile);
    // fprintf(stderr,"... Outputting list of unmatched fluid nodes .... %s\n", outname);
    fprintf(stderr,"... Outputting BC-Data file ........................ %s\n", outname);
    FILE* nonMatchOutFluid = fopen(outname, "w");
    fprintf(nonMatchOutFluid, "SDBoundary NonMatched using FluidNodes\n");
    sprintf(outname, "%s.nonMatch.top", fn.outFile);
    FILE* nonMatchOutTop = fopen(outname, "w");
    fprintf(nonMatchOutTop, "Nodes NonMatchedNodes\n");
    fprintf(stderr,"... Outputting unmatched fluid nodes xpost file .... %s\n", outname);
    for (i=0; i<count; i++) {
      if(!coords[interfNode[i]].success) {
        fprintf(nonMatchOutFluid, "%8d  %d  %f\n", interfNode[i]+1,1,0.0);
        fprintf(nonMatchOutFluid, "%8d  %d  %f\n", interfNode[i]+1,2,0.0);
        fprintf(nonMatchOutFluid, "%8d  %d  %f\n", interfNode[i]+1,3,0.0);
        fprintf(nonMatchOutTop,"%8d %g %g %g\n", ++nonMatchNum, fluidDom->nodes[ interfNode[i] ][0],
		fluidDom->nodes[ interfNode[i] ][1], fluidDom->nodes[ interfNode[i] ][2]);
        fprintf(nonMatchOutTop,"%8d %g %g %g\n", ++nonMatchNum, fluidDom->nodes[ interfNode[i] ][0],
		fluidDom->nodes[ interfNode[i] ][1], fluidDom->nodes[ interfNode[i] ][2]);
      }
    }
    
    fclose(nonMatchOutFluid); 
    
    fprintf(nonMatchOutTop, "Elements NonMatchedElements using NonMatchedNodes\n");
    nonMatchNum=nonMatchNum/2;
    for (i=1;i<=nonMatchNum;i++)
      fprintf(nonMatchOutTop, "%8d 1 %8d %8d\n", i, 2*i-1, 2*i);
    fclose(nonMatchOutTop);
  }
  
  
/* for debugging
  sprintf(outname, "%s.fnodes1.top", fn.outFile);
  FILE* matchOut1 = fopen(outname, "w");
  fprintf(matchOut1, "Nodes FluidInterfaceNodes1\n");
  numOut=0;
  for (i=0; i<count; i++) {
    if (coords[ interfNode[i] ].success == 1) {
      fprintf(matchOut1, "%d %g %g %g\n", ++numOut, fluidDom->nodes[ interfNode[i] ][0],
              fluidDom->nodes[ interfNode[i] ][1], fluidDom->nodes[ interfNode[i] ][2]);
    }
  }
  fprintf(matchOut1, "Elements FluidInterface1 using FluidInterfaceNodes1\n");
  for (i=1;i<=numOut;i++)
    fprintf(matchOut1, "%d 1 %d %d\n", i, i, i);
  fclose(matchOut1);

  sprintf(outname, "%s.fnodes2.top", fn.outFile);
  FILE* matchOut2 = fopen(outname, "w");
  fprintf(matchOut2, "Nodes FluidInterfaceNodes2\n");
  numOut=0;
  for (i=0; i<count; i++) {
    if (coords[ interfNode[i] ].success != 1) {
      fprintf(matchOut2, "%d %g %g %g\n", ++numOut, fluidDom->nodes[ interfNode[i] ][0],
              fluidDom->nodes[ interfNode[i] ][1], fluidDom->nodes[ interfNode[i] ][2]);
    }
  }
  fprintf(matchOut2, "Elements FluidInterface2 using FluidInterfaceNodes2\n");
  for (i=1;i<=numOut;i++)
    fprintf(matchOut2, "%d 1 %d %d\n", i, i, i);
  fclose(matchOut2);

  sprintf(outname, "%s.scoords.top", fn.outFile);
  FILE* matchOut3 = fopen(outname, "w");
  fprintf(matchOut3, "Nodes StructMatchingPoints\n");
  numOut=0;
  for (i=0; i<count; i++) {
    if (coords[ interfNode[i] ].success == 1) {
      fprintf(matchOut3, "%d %g %g %g\n", ++numOut, coords[ interfNode[i] ].point[0],
              coords[ interfNode[i] ].point[1], coords[ interfNode[i] ].point[2]);
    }
  }
  fprintf(matchOut3, "Elements StructMatch using StructMatchingPoints\n");
  for (i=1;i<=numOut;i++)
    fprintf(matchOut3, "%d 1 %d %d\n", i, i, i);
  fclose(matchOut3);
*/

  // Read Fluid Decomposition file if available
  decomp->getDecomp(fn.decF);
  if (fn.decF!=NULL)
    fprintf(stderr, "...Fluid decomposition read from: %s\n", fn.decF);
  else {
    if (type == 0)
      fprintf(stderr, "...No fluid decomposition file: Assuming 1 fluid subdomain\n");
  }

  // Form the subdomains in decomp file
  decomp->makeSubD(noOverlap);
  if (type == 0)
    fprintf (stderr, "... Fluid subdomains constructed\n");

  // now the interfaces are found for each subdomain
  decomp->getDomainConnect();
  if (type == 0)
    fprintf (stderr, "... Fluid subdomains' interfaces found\n");

  // a sub to node connectivity is returned so that we can output
  // the flu+ files and the MATCHER file
  Connectivity* subToNode = decomp->getSubToNode();

  //also the number of fluid sub's is returned
  numFluidSub = decomp->getNumSub();

  // output the flu- and flu.glob files if running for old fluid code
  if(type == 0) decomp->print();

  /*
    the preceeding functions are defined in Decomp3D.{h,C}
    and Decomp2D.{h,C} depending on the dimension of the problem
    theses functions must be called in this order as each depends
    on information supplied by the prior functions
    
    
    we now construct a connectivity between the structural subs
    and the fluid subs across the matches (ie this describes
    what struct subs have elements matched to nodes of what
    fluid subs)
    the reverse connectivity is also created
  */
  
  int** numMatch=new int*[numFluidSub];
  count=0;
  int* cx=new int[numFluidSub+1];
  int* connect=new int[numFluidSub*numStructSub];
  int* touches=new int[numStructSub];
  
  for (i=0; i<numFluidSub; i++)  {
    cx[i]=count;
    numMatch[i] = new int[numStructSub];
    
    for (j=0;j<numStructSub;j++){
      touches[j]=0;
      numMatch[i][j]=0;
    }
    
    for (j=0; j<subToNode->num(i); j++){
      nnum = (*subToNode)[i][j];

      if (coords[nnum].success==1) {
	snum = coords[nnum].domNum;
 	numMatch[i][snum]++;

	if (touches[snum]==0) {
  	  touches[snum]=1;
	  connect[count++]=snum;
	}
      }
    }
  }

  cx[numFluidSub]=count;
  
  // create fluidNode to structure element connectivity
  Connectivity *fluidToStruct = new Connectivity(numFluidSub,cx,connect);
  
  // creates struct element to fluid node connectivity
  Connectivity *structToFluid = fluidToStruct->reverse();
  
  // Create file name prefix for new fluid code
  int length;
  char *lastDot = strrchr(fn.fluid,'.');
  
  if (!lastDot)
    length = strlen(fn.fluid);
  else
    length = lastDot - fn.fluid;
  
  char *outputFile = new char[length+12];
  strncpy(outputFile, fn.fluid, length);

  // Output of fluid data into fluid.match0#
  int maxLength;
  if(type == 0) {
    strcpy(outFile,"flu+00000");
    //here come the flu+ files, the first part is a semi-clever
    //idea to increment the file names in a loop
    for(i=0;i<numFluidSub;i++){

      for(j=8;j>3;j--){
	if(outFile[j]!='9'){
   	  outFile[j]++;
	  break;
	}
        else
	  outFile[j]='0';
      }
      ofstream fluOut(outFile,ios::out);

      //ok 1st line, the number struct subs connected to this fluid sub
      fluOut<<fluidToStruct->num(i)<<endl;
      maxLength=0;

      for(j=0;j<fluidToStruct->num(i);j++){
	snum=(*fluidToStruct)[i][j];

	//the number of the first struct sub that is connected 
	fluOut<<snum+1<<' ';

	//and the number of matches between the two
	fluOut<<numMatch[i][snum]<<endl;

	//calc the max number of matches over all struct subs for this flu sub
	maxLength=(maxLength<numMatch[i][j])?numMatch[i][j]:maxLength;

	//output the node numbers (in local coords) that match 
	//this struct sub and are in this flu sub
	for(k=0;k<subToNode->num(i);k++){
 	  nnum=(*subToNode)[i][k];
	  if (coords[nnum].success==1 && coords[nnum].domNum==snum)
	    fluOut<<k+1<<endl;
	}
      }

      //the zeroes have to do with Gauss points, which we aren't using
      //and maxLength again is the max message length between this flu sub
      //and all struct subs
      fluOut<<0<<endl; 
      fluOut<<maxLength<<endl;
      fluOut<<0<<endl;

      // Now output the 'gap vector'
      for(j=0; j<fluidToStruct->num(i); j++)  {
        for(k=0;k<subToNode->num(i);k++){
          nnum=(*subToNode)[i][k];
          if (coords[nnum].success==1 && coords[nnum].domNum==snum)  {
            Point3D gap = fluidDom->nodes[nnum]-coords[nnum].point;
            fluOut<< gap<<endl;
          }
        }
      }

    //close this bad boy down 
    fluOut.close();
    fprintf(stderr,"... Outputting fluid matcher file ............. %s\n", outFile);
    }
  } 

  else { // New fluid code output
    for(i=0;i<numFluidSub;i++){
      if (numFluidSub == 1)
	sprintf(outname, "%s.match.fluid", fn.outFile);
      else
	sprintf(outname, "%s.match.fluid%d", fn.outFile, i);
      FILE* fluOut = fopen(outname, "w");

      fprintf(stderr,"... Outputting fluid matcher file .................. %s \n", outname);
      //ok 1st line, the number struct subs connected to this fluid sub
      fprintf(fluOut, "%d\n", fluidToStruct->num(i));
      maxLength=0;

      for(j=0; j<fluidToStruct->num(i); j++)  {
        //the number of the first struct sub that is connected 
        //and the number of matches between the two
        snum=(*fluidToStruct)[i][j];
	fprintf(fluOut, "%d %d\n", snum+1, numMatch[i][snum]);
	        
	//output the node numbers (in local coords) that match 
        //this struct sub and are in this flu sub
	for(k=0; k<subToNode->num(i); k++)  {
      	  nnum=(*subToNode)[i][k];
    	  if (coords[nnum].success==1 && coords[nnum].domNum==snum)
	    fprintf(fluOut, "%d\n", k+1);
        }

        // Now output the 'gap vector'
        for (k=0; k<subToNode->num(i); k++)  {
     	  nnum=(*subToNode)[i][k];
	  if (coords[nnum].success==1 && coords[nnum].domNum==snum){
	    Point3D gap = fluidDom->nodes[nnum]-coords[nnum].point;
	    fprintf(fluOut, "%e %e %e\n", gap[0], gap[1], gap[2]);
	  }
        }
      }  
      fclose(fluOut);
    }
  }

  // Output structure data into struct.match
  // create outfile prefix name
  lastDot = strrchr(fn.structure, '.');

  if (!lastDot)
    length = strlen(fn.structure);
  else
    length = lastDot - fn.structure;
     
  char *fluidmatchFile = new char[length+9]; 
  strncpy(fluidmatchFile, fn.structure, length);

  for (i=0; i<numStructSub; i++){

    if (numStructSub == 1)
      sprintf(outname, "%s.match.fem", fn.outFile);
    else
      sprintf(outname, "%s.match.fem%d", fn.outFile, i);
    FILE* MATCHOut = fopen(outname, "w");

    fprintf(stderr,"... Outputting structure matcher file .............. %s\n", outname);
    //number of this struct sub
    fprintf(MATCHOut, "SUBD %d\n", i+1);

    //no gauss points so no received points
    fprintf(MATCHOut, "RCVF 0\n");

    //number of matched elems (sent elements to fluid)
    fprintf(MATCHOut, "SNDF %d\n", structToFluid->num(i));

    for (j=0; j<structToFluid->num(i); j++) {
      fnum=(*structToFluid)[i][j];

      //for each connected fluid sub output this flu sub's and
      //the number of points that were matched in it 
      fprintf(MATCHOut, "%d %d\n", fnum+1, numMatch[fnum][i]);

      //output the element number and natural coords of
      //all successfull matches between these 2 subs 
      for (k=0; k<subToNode->num(fnum); k++){
        nnum=(*subToNode)[fnum][k];
      	if (coords[nnum].success==1 && coords[nnum].domNum==i) {
          Point3D gap = fluidDom->nodes[nnum]-coords[nnum].point;
          fprintf(MATCHOut, "%d %e %e %e %e %e\n", coords[nnum].elNum+1, 
		  coords[nnum].xi, coords[nnum].eta, gap[0], gap[1], gap[2]);
       	}
      }
    }
    fprintf(MATCHOut, "END SUBD\n");
    fprintf(MATCHOut, "END\n");
    fclose(MATCHOut);
  }
  //MATCHOut<<"END"<<endl;

  //reset the output (prob. not necesarry but it can't hurt)
  //MATCHOut.setf(oldbase,ios::fixed);

  fprintf(stderr,"---------------------------------------------------------\n");
  fprintf(stderr,"... Matching CPU time = %f s. \n", totalParallelTime);
  progTime = progTimer.getTime() - progTime;
  fprintf(stderr,"... Total CPU time = %f s. \n", progTime);
  return 0;
}

//----------------------------------------------------------------------

/* 
Possible Arguments
 -d dim sets the dimension of the problem
 -t (opt) tol sets the tolerance to tol
 -n code uses new version of matcher
 -p number of processors (threads)
 -l number of levels of tolerance iteration 
 -e maximum acceptable distance between matched fluid-structure pair

 TOP DOMDEC format is available
 -f (opt) specifies fluid decomposition file
 -s (opt) specifies a structural decomposition
 -b (opt) tells matcher that a file containing CNBC info is available  
 -a (opt) tells matcher that a struct cntrl file is available
 -o (opt) turns off the overlap in the 2D code

*/

int readCommandLine(int argc, char **argv, FileNames &fn, int &decomposeFlag, 
		    int &bcFlag, int &noOverlap, int &numThreads, double &tol, double &maxdist,
                    int &numtol)  
{

  fn.outFile = "OUTPUT";

  // First cmd line argument is the fluid mesh file
  fn.fluid = argv[1];

  //Check if 2nd argument is the structure input file
  int iArg;
  if (argc > 2 && argv[2][0] != '-') {
    iArg = 3;
    fn.structure = argv[2];
  }
  else  {
    // Only decomposing fluid mesh file
    iArg = 2;
    decomposeFlag = 1;
  }
  
  int type = 1;

  // Parse switch arguments
  while (iArg < argc) {

    if (strcmp(argv[iArg], "-t") == 0)  { 
      tol = atof(argv[iArg+1]);
      if (tol <= 0.0)  {
        fprintf(stderr,"***ERROR: Tolerance, %f, must be a postive number\n", tol);
        exit(1);
      }
    }
    if (strcmp(argv[iArg], "-e") == 0)  {
      maxdist = atof(argv[iArg+1]);
      if (maxdist <= 0.0)  {
        fprintf(stderr,"***ERROR: Maximum acceptable distance, %f, must be a postive number\n", maxdist);
        exit(1);
      }
    }
    if (strcmp(argv[iArg], "-l") == 0)  {
      numtol = atoi(argv[iArg+1]);
      if (numtol <= 0)  {
        fprintf(stderr,"***ERROR: Number of tolerance levels, %d, must be a postive number\n", numtol);
        exit(1);
      }
    }

    if (strcmp(argv[iArg], "-a") == 0)  { 
      fn.control = argv[iArg+1]; 
    }
    if (strcmp(argv[iArg], "-d") == 0)  { 
      dimension = atoi(argv[iArg+1])-1; 
      if (dimension != 1 && dimension != 2)  {
        fprintf(stderr, "*** ERROR: Invalid dimension specified: %d\n", 
			dimension);
        fprintf(stderr, "... Aborting\n");
        exit(1);
      }
    }
/* new is default
    if (strncmp(argv[iArg], "-new", 2) == 0)  { 
      fprintf(stderr,"... Executing new version of Matcher\n");
      type++;
    }
    type++; // does what -new did
*/

    if (strcmp(argv[iArg], "-p") == 0)  { 
      numThreads = atoi(argv[iArg+1]);
#ifndef _OPENMP
      if(numThreads>1) { // HB (09/26/05) warning message if more than 1 thread requested when running a non-OpenMP executable
        fprintf(stderr,"*** Warning: not an OpenMP executable (should be compiled with an OpenMP compliant compiler & flag _OPENMP enabled).\n");

        fprintf(stderr,"*** Warning: number of thread set to 1.\n");
        numThreads = 1;
      }
#endif
    }
    if (strcmp(argv[iArg], "-f") == 0)  { 
      fn.decF=argv[iArg+1]; 
    }
    if (strcmp(argv[iArg], "-s") == 0)  { 
      fn.decS=argv[iArg+1]; 
    }
    if (strcmp(argv[iArg], "-b") == 0)  { 
      fn.bcFile=argv[iArg+1]; 
      bcFlag=1;
    }
    if (strcmp(argv[iArg], "-n") == 0)  {
      fn.matchedNodes=argv[iArg+1];
    }
    if (strcmp(argv[iArg], "-m") == 0)  {
      fn.matchedElements=argv[iArg+1];
    }
    if (strncmp(argv[iArg], "-overlap", 3) == 0)  { 
      fprintf(stderr,"... Overlap disabled\n");
      noOverlap = 1;
    }
    if (strcmp(argv[iArg], "-beam") == 0) {
      beamFlag = true;
    }
    if (strncmp(argv[iArg], "-output", 3) == 0)  { 
      fn.outFile = argv[iArg+1]; 
    }
    iArg++;
  }

  // Check if Overlapping was selected with 3D case
  if (dimension == 2 && noOverlap == 1)
    fprintf(stderr, "*** WARNING: -o flag ignored in the 3D case\n");
	
  // Check if fluid control file is specified in 2D case
  if(dimension == 1 && bcFlag == 1)  {
    fprintf(stderr, "*** ERROR: The -b flag is not supported for 2D meshes\n");
    fprintf(stderr, "... Exiting\n");
    exit(1);
  }

  if (decomposeFlag == 1 && fn.control == NULL)  {
    fprintf(stderr, "Incorrect usage\n");
    fprintf(stderr, "Usage: matcher <fluid mesh file>\n");
    fprintf(stderr, "               <structure mesh file>\n");
    fprintf(stderr, "               [-b <fluid.control file>]\n");
    fprintf(stderr, "               [-a <structure.control file>]\n");
    fprintf(stderr, "               [-t <geometric tolerance>]\n");
    fprintf(stderr, "               [-e <distance tolerance>]\n");
    fprintf(stderr, "               [-m <fluid.match file>]\n");
    fprintf(stderr, "               [-n <structure.match file>]\n");
    fprintf(stderr, "               [-beam <beam elements>\n");
    fprintf(stderr, "               [-l <number of levels for the geometric tolerance>]\n");
    fprintf(stderr, "               [-output <output prefix>]\n");
    fprintf(stderr, "               [-p <number of threads>]\n");
    exit(1);
  }

  return type;
}
