#include"Quadralateral3D.h"

//#define QUAD4_TRIANGLE_APPROX_4 //approximate search by subdiving the quad into 4 triangles using the geometric center of gravity of the quad el.
#define QUAD4_TRIANGLE_APPROX_2 //approximate search by subdiving the quad into 2 triangles follwing the FEM code convenvtion (see FEM el. 88 or 2020)

#define QUAD4_FLAT_TOL 1.E-8 // (absolute) tolerance for detecting a flat quad4 el. using ||d1|| as criterion: 
                             // ||d1|| < QUAD4_FLAT_TOL -> Quad4 is considered as flat ...

Quadralateral3D::Quadralateral3D(int inSize, int* inNodes)
{
  assert(inSize==4);
  nodes[0]=inNodes[0];
  nodes[1]=inNodes[1];
  nodes[2]=inNodes[2];
  nodes[3]=inNodes[3];
}
int
Quadralateral3D::getNode(int position)
{
  assert(0<=position);
  assert(position<=3);
  return(nodes[position]);
}

// solve a 2x2 linear system using given rotation if selected
void solveAxb2x2(double A[2][2], double b[2], double x[2], bool givenRot=false, double tol=1.E-4)
{
  if(givenRot) { // use givens rotation to solve Ax=b
    double c,s,t;
    if(fabs(A[1][0])<tol) { c=1.0; s=0.0; }
    else {
      if(fabs(A[1][0])>fabs(A[0][0])){
         t=-A[0][0]/A[1][0];
         s=1.0/(sqrt(1+t*t));
         c=s*t;
      }else{
         t=-A[1][0]/A[0][0];
         c=1.0/(sqrt(1+t*t));
         s=c*t;
       }
     }
     double t1=A[0][1];
     A[0][0]=c*A[0][0]-s*A[1][0];
     A[0][1]=c*A[0][1]-s*A[1][1];
     A[1][1]=s*t1+c*A[1][1];
     t1=b[0];
     b[0]=c*t1-s*b[1];
     b[1]=s*t1+c*b[1];
     x[1]=b[1]/A[1][1];
     x[0]=(b[0]-x[1]*A[0][1])/A[0][0];
  } else { // standard method
    double det = A[0][0]*A[1][1] - A[0][1]*A[1][0];
#ifdef HB_WARNING
    if(fabs(det)<1.E-12) { fprintf(stderr," *** WARNING: in solveAxb2x2: matrix nearly singular. det = %e\n", det);}
#endif
    x[0] = ( A[1][1]*b[0] - A[0][1]*b[1])/det;
    x[1] = (-A[1][0]*b[0] + A[0][0]*b[1])/det;
  }
}

#if defined(QUAD4_TRIANGLE_APPROX_2) || defined(QUAD4_TRIANGLE_APPROX_4)
// return true if projected point is "in" the triangle using the three tolerances given in array tols (i.e. one for each edge), otherwise false.
bool isInTriangle(const Point3D& X, Point3D& X1, Point3D& X2, Point3D& X3, MatchInfo& info, double tols[3])
{ 
  Point3D X12 = X2 - X1;
  Point3D X13 = X3 - X1;
  Point3D X1X = X  - X1;
  
  double A[2][2],b[2],x[2];
  A[0][0]= dotp(X12,X12);
  A[1][0]= A[0][1] = dotp(X12,X13);
  A[1][1]= dotp(X13,X13);
  b[0]   = dotp(X12,X1X);
  b[1]   = dotp(X13,X1X);

  // uses givens rotation to solve Ax=b
  solveAxb2x2(A,b, x, true, 1.E-4);
  info.xi = x[0];
  info.eta= x[1];
  info.success=0;
  if((-tols[2]<=info.xi)&&(-tols[0]<=info.eta)&&(-2.*tols[1]<=(1-info.eta-info.xi))){
    info.success = 1;
    info.point   = X1 + info.xi*X12 + info.eta*X13; // triangular mapping
    return(true);
  }
  return(false);
}
#endif

#if defined(QUAD4_TRIANGLE_APPROX_4) // Subdivide the quad4 into 4 triangles
void
Quadralateral3D::match(ResizeArray<Point3D*>& allPoints,
const Point3D& matchPoint, const double tol, MatchInfo& info)const 
{
  info.success = 0;

  Point3D x1=*(allPoints[nodes[0]]);
  Point3D x2=*(allPoints[nodes[1]]);
  Point3D x3=*(allPoints[nodes[2]]);
  Point3D x4=*(allPoints[nodes[3]]);
  Point3D d1=x1+x3-x4-x2;
  Point3D d2=x2-x1;
  Point3D d3=x4-x1;

  if(norm(d1)<QUAD4_FLAT_TOL){ // Quad is flat ...
#ifdef QUAD4_DEBUG
    fprintf(stderr," *** In Quadralateral3D::match (4 triangles approx): quad is flat, ||d1|| = %e\n",norm(d1));
#endif    
    double A[2][2],b[2],x[2];
    Point3D mmp=matchPoint-x1;
    A[0][0]= dotp(d2,d2);
    A[1][0]= A[0][1] = dotp(d2,d3);
    A[1][1]= dotp(d3,d3);
    b[0]   = dotp(mmp,d2);
    b[1]   = dotp(mmp,d3);
    // uses givens rotation to solve Ax=b
    solveAxb2x2(A,b, x, true, 1.E-4);
    info.xi = x[0];
    info.eta= x[1];
    if((-tol<=info.xi)&&(info.xi<=1.0+tol)){
      if((-tol<=info.eta)&&(info.eta<=1.0+tol)){
        info.success= 1;
        info.point  = x1 + (info.eta*info.xi)*d1 + info.xi*d2 + info.eta*d3;
        info.dist   = norm(info.point-matchPoint);
        return;
      }
    }
  } else { // Non flat case: approximate by subdividing the Quad into 4 triangles (flat)
#ifdef QUAD4_DEBUG
     fprintf(stderr," *** In Quadralateral3D::match (4 triangles approx): quad is NOT flat, ||d1|| = %e\n",norm(d1));
#endif
     Point3D xc = 0.25*(x1+x2+x3+x4);
     double tols[3] = {0.0,tol,0.0};
       
     // first triangle (xc,x1,x2)
     if(isInTriangle(matchPoint, xc, x1, x2, info, tols)) {
       double x = 0.5*(1.-info.xi+info.eta);
       double y = 0.5*(1.-info.xi-info.eta);
       info.xi = x; 
       info.eta= y;
       info.success = 1;
       info.point   = x1 + (info.xi*info.eta)*d1 + info.xi*d2 + info.eta*d3; // point on the true quad geometry
       info.dist    = norm(info.point-matchPoint);
       return; 
     }

     // second triangle (xc,x2,x3)
     if(isInTriangle(matchPoint, xc, x2, x3, info, tols)) {
       double x = 0.5*(1.+info.xi+info.eta);
       double y = 0.5*(1.-info.xi+info.eta);
       info.xi = x;
       info.eta= y;
       info.success = 1;
       info.point   = x1 + (info.xi*info.eta)*d1 + info.xi*d2 + info.eta*d3; // point on the true quad geometry
       info.dist    = norm(info.point-matchPoint);
       return;
     }
 
     // third triangle (xc,x3,x4)
     if(isInTriangle(matchPoint, xc, x3, x4, info, tols)) {
       double x = 0.5*(1.+info.xi-info.eta);
       double y = 0.5*(1.+info.xi+info.eta);
       info.xi = x;
       info.eta= y;
       info.success = 1;
       info.point   = x1 + (info.xi*info.eta)*d1 + info.xi*d2 + info.eta*d3; // point on the true quad geometry
       info.dist    = norm(info.point-matchPoint);
       return;
     }

     // fourth triangle (xc,x4,x1)
     if(isInTriangle(matchPoint, xc, x4, x1, info, tols)) {
       double x = 0.5*(1.-info.xi-info.eta);
       double y = 0.5*(1.+info.xi-info.eta);
       info.xi = x;
       info.eta= y;
       info.success = 1;
       info.point   = x1 + (info.xi*info.eta)*d1 + info.xi*d2 + info.eta*d3; // point on the true quad geometry
       info.dist    = norm(info.point-matchPoint);
       return;
     }
  }
#ifdef QUAD4_DEBUG
  fprintf(stderr," ***In Quadralateral3D::match (4 triangles approx): NOT MATCHED ...\n");
#endif
}
#elif defined(QUAD4_TRIANGLE_APPROX_2) // Subdivide the quad4 into 2 triangles following the FEM convenvtion (see FEM el. 88 or 2020)
void
Quadralateral3D::match(ResizeArray<Point3D*>& allPoints,
const Point3D& matchPoint, const double tol, MatchInfo& info)const 
{
  info.success = 0;

  Point3D x1=*(allPoints[nodes[0]]);
  Point3D x2=*(allPoints[nodes[1]]);
  Point3D x3=*(allPoints[nodes[2]]);
  Point3D x4=*(allPoints[nodes[3]]);
  Point3D d1=x1+x3-x4-x2;
  Point3D d2=x2-x1;
  Point3D d3=x4-x1;

  if(norm(d1)<QUAD4_FLAT_TOL){ // Quad is flat ...
#ifdef QUAD4_DEBUG
    fprintf(stderr," *** In Quadralateral3D::match (2 triangles approx): quad is flat, ||d1|| = %e\n",norm(d1));
#endif    
    double A[2][2],b[2],x[2];
    Point3D mmp=matchPoint-x1;
    A[0][0]= dotp(d2,d2);
    A[1][0]= A[0][1] = dotp(d2,d3);
    A[1][1]= dotp(d3,d3);
    b[0]   = dotp(mmp,d2);
    b[1]   = dotp(mmp,d3);
    // uses givens rotation to solve Ax=b
    solveAxb2x2(A,b, x, true, 1.E-4);
    info.xi = x[0];
    info.eta= x[1];
    if((-tol<=info.xi)&&(info.xi<=1.0+tol)){
      if((-tol<=info.eta)&&(info.eta<=1.0+tol)){
        info.success= 1;
        info.point  = x1 + (info.eta*info.xi)*d1 + info.xi*d2 + info.eta*d3;
        info.dist   = norm(info.point-matchPoint);
        return;
      }
    }
  } else { // Non flat case: approximate by subdividing the Quad into 2 triangles (flat)
#ifdef QUAD4_DEBUG
     fprintf(stderr," *** In Quadralateral3D::match (2 triangles approx): quad is NOT flat, ||d1|| = %e\n",norm(d1));
#endif
     double tols[3] = {tol,0.0,tol};
       
     // first triangle (x2,x3,x1)
     if(isInTriangle(matchPoint, x2, x3, x1, info, tols)) {
       double x = 1.-info.eta;
       double y = info.xi;
       info.xi = x; 
       info.eta= y;
       info.success = 1;
       info.point   = x1 + (info.xi*info.eta)*d1 + info.xi*d2 + info.eta*d3; // point on the true quad geometry
       info.dist    = norm(info.point-matchPoint);
       return; 
     }

     // second triangle (x4,x1,x3)
     if(isInTriangle(matchPoint, x4, x1, x3, info, tols)) {
       double x = info.eta;
       double y = 1.-info.xi;
       info.xi = x;
       info.eta= y;
       info.success = 1;
       info.point   = x1 + (info.xi*info.eta)*d1 + info.xi*d2 + info.eta*d3; // point on the true quad geometry
       info.dist    = norm(info.point-matchPoint);
       return;
     }
  }
#ifdef QUAD4_DEBUG
  fprintf(stderr," ***In Quadralateral3D::match (2 triangles approx): NOT MATCHED ...\n");
#endif
}

#else // full nonlinear resolution based on a minization formulation 
      // solved by a Newton method with line search

double lineSearch2(double xx[2], double d[2], double R[2], 
		  Point3D& d1, Point3D& d2, Point3D& d3,
		  Point3D& X1, Point3D& X)
{
  int jMax = 3;
  double beta = 1.E-4;
  double eta  = 0.1;
  double alpha = 1.;
  double M0 = 0.5*(R[0]*R[0]+R[1]*R[1]);
  double dM0= -2.*M0;
  double x, y;
  Point3D DX, Nx, Ny;

  for(int j=0; j<jMax; j++){
    // new unknow, residual & merit function
    x = xx[0] + alpha*d[0];
    y = xx[1] + alpha*d[1];

    DX[0] = X[0] - (X1[0] + x*d2[0] + y*d3[0] + (x*y)*d1[0]);
    DX[1] = X[1] - (X1[1] + x*d2[1] + y*d3[1] + (x*y)*d1[1]);
    DX[2] = X[2] - (X1[2] + x*d2[2] + y*d3[2] + (x*y)*d1[2]);
    Nx[0] = d2[0] + y*d1[0]; Nx[1] = d2[1] + y*d1[1]; Nx[2] = d2[2] + y*d1[2];
    Ny[0] = d3[0] + x*d1[0]; Ny[1] = d3[1] + x*d1[1]; Ny[2] = d3[2] + x*d1[2];
    R[0] = dotp(Nx,DX);
    R[1] = dotp(Ny,DX);
      
    double M = 0.5*(R[0]*R[0]+R[1]*R[1]);
   
    // check Goldstein's condition
    if(M<=(1.-2.*beta*alpha)*M0) { xx[0] = x; xx[1] = y; return(alpha); }
   
    double c = -0.5*alpha*alpha*dM0/(M-M0-alpha*dM0);
    if(c>eta*alpha) alpha = c;
    else alpha *= eta;
  }
  // check maxinum number of quadratic fit
  //fprintf(stderr," *** WARNING: in lineSearch: maxinum number of curve fitting reached. alpha = %e\n",alpha);
  xx[0] = x; xx[1] = y; 
  return(alpha);      
}

void
Quadralateral3D::match(ResizeArray<Point3D*>& allPoints,
const Point3D& matchPoint, const double tol, MatchInfo& info)const {

	Point3D x1=*(allPoints[nodes[0]]);
	Point3D	x2=*(allPoints[nodes[1]]);
	Point3D x3=*(allPoints[nodes[2]]);
	Point3D x4=*(allPoints[nodes[3]]);
	Point3D d1=x1+x3-x4-x2;
	Point3D d2=x2-x1;
	Point3D d3=x4-x1;
	
	// Initial guess assuming d1 = 0
        Point3D xc  = 0.25*(x1+x2+x3+x4);
        Point3D x12 = 0.5*(x1+x2);
        Point3D x23 = 0.5*(x2+x3);
        Point3D x34 = 0.5*(x3+x4);
        Point3D x41 = 0.5*(x4+x1);
	Point3D T1  = 0.5*(x23 - x41);
	Point3D T2  = 0.5*(x34 - x12);
	Point3D N   = crossp(T1,T2);
	Point3D n   = N/norm(N);
	
	double alpha=dotp(matchPoint-xc,n);
        info.point=matchPoint-alpha*n;
        Point3D mmp=info.point-xc;			
	double c,s,t;
	double A[3][3],b[3];
	A[0][0]=dotp(T1,T1);
	A[1][0]=A[0][1]=dotp(T1,T2);
        A[1][1]=dotp(T2,T2);
        b[0]=dotp(mmp,T1);
        b[1]=dotp(mmp,T2);
        double det = A[0][0]*A[1][1] - A[0][1]*A[1][0];
	
        double x = ( A[1][1]*b[0] - A[0][1]*b[1])/det;
        double y = (-A[1][0]*b[0] + A[0][0]*b[1])/det;
        x = 0.5*(x+1.); y = 0.5*(y+1.);

#define SPEED_UP
#ifdef SPEED_UP
        if((fabs(x)>5.)||(fabs(y)>5.)) { info.success = 0; return; } //HB: to speed up things: do not try to "polish" 
                                                                       //    the solution if it already out of the element
#endif
        Point3D X = matchPoint; //just to avoid the const ...
        Point3D X1= x1;
	
	Point3D DX = matchPoint - (x1 + x*d2 + y*d3 + (x*y)*d1);
        Point3D Nx = d2 + y*d1;
        Point3D Ny = d3 + x*d1;
        int itr = 0;
	int itrMax= 10;
        double dx[2] = {0.0,0.0};
        double R[2] = {dotp(Nx,DX), dotp(Ny,DX)};
	double nR = sqrt(R[0]*R[0]+R[1]*R[1]);
        double nR0 = nR;
	 
        bool useLineSearch = true;
        bool thirdOrder    = true;
	double rtol = 1.E-8;
	double atol = 1.E-12;
	
	//fprintf(stderr," ******************************************* \n");
	while((nR> rtol*nR0 + atol)&(itr<itrMax)) {
	   itr++;
	   //fprintf(stderr," -> Iteration %2d, x = %e, y= %e, ||R|| = %e\n",itr,x,y,nR);

	   A[0][0] = -dotp(Nx,Nx); 
	   A[0][1] = A[1][0]= dotp(d1,DX)-dotp(Nx,Ny);
	   A[1][1] = -dotp(Ny,Ny); 

	   b[0] = -R[0]; 
	   b[1] = -R[1]; 

           det = A[0][0]*A[1][1] - A[0][1]*A[1][0];
	   //fprintf(stderr," -> det = %e\n",det); 
           dx[0] = ( A[1][1]*b[0] - A[0][1]*b[1])/det;
           dx[1] = (-A[1][0]*b[0] + A[0][0]*b[1])/det;

           if(thirdOrder) {
	     double xmid = x + 0.5*dx[0];
	     double ymid = y + 0.5*dx[1];
	     Nx[0] = d2[0] + ymid*d1[0]; Nx[1] = d2[1] + ymid*d1[1]; Nx[2] = d2[2] + ymid*d1[2];
             Ny[0] = d3[0] + xmid*d1[0]; Ny[1] = d3[1] + xmid*d1[1]; Ny[2] = d3[2] + xmid*d1[2];
             DX[0] = X[0] - (X1[0] + xmid*d2[0] + ymid*d3[0] + (xmid*ymid)*d1[0]);
             DX[1] = X[1] - (X1[1] + xmid*d2[1] + ymid*d3[1] + (xmid*ymid)*d1[1]);
             DX[2] = X[2] - (X1[2] + xmid*d2[2] + ymid*d3[2] + (xmid*ymid)*d1[2]);
	 
	     A[0][0] = -dotp(Nx,Nx); 
	     A[0][1] = A[1][0]= dotp(d1,DX)-dotp(Nx,Ny);
	     A[1][1] = -dotp(Ny,Ny); 

             det = A[0][0]*A[1][1] - A[0][1]*A[1][0];
	   
             dx[0] = ( A[1][1]*b[0] - A[0][1]*b[1])/det;
             dx[1] = (-A[1][0]*b[0] + A[0][0]*b[1])/det;
	   }
	   if(useLineSearch) {
	     double xx[2] = {x,y};
             double ls = lineSearch2(xx, dx,R, d1, d2, d3, X1, X);
	     //fprintf(stderr," -> step size = %e\n", ls); 
	     x = xx[0]; y = xx[1]; 
	   } else {		  
	     x = x + dx[0];
	     y = y + dx[1];
	   }  

           DX[0]= X[0] - (X1[0] + x*d2[0] + y*d3[0] + (x*y)*d1[0]);
           DX[1]= X[1] - (X1[1] + x*d2[1] + y*d3[1] + (x*y)*d1[1]);
           DX[2]= X[2] - (X1[2] + x*d2[2] + y*d3[2] + (x*y)*d1[2]);
           Nx[0]= d2[0] + y*d1[0]; Nx[1] = d2[1] + y*d1[1]; Nx[2] = d2[2] + y*d1[2];
           Ny[0]= d3[0] + x*d1[0]; Ny[1] = d3[1] + x*d1[1]; Ny[2] = d3[2] + x*d1[2];
	   R[0] = dotp(Nx,DX);
	   R[1] = dotp(Ny,DX);
	   nR   = sqrt(R[0]*R[0]+R[1]*R[1]);
	}  
	//if(itr==itrMax) 
	//   fprintf(stderr," *** WARNING: maximun number of iterations reached. Last ||R|| = %e\n",nR);

        info.eta=y;
	info.xi=x;
	info.success=0;
  	if((-tol<=info.xi)&&(info.xi<=1.0+tol)){
		if((-tol<=info.eta)&&(info.eta<=1.0+tol)){
                	info.point=x1+info.eta*info.xi*d1+info.xi*d2+info.eta*d3;
			info.dist=norm(info.point-matchPoint);
			info.success=1;
		}
	}

}
#endif

void
Quadralateral3D::getLocalCoordinates(int i, double &xi, double &eta)
{
  // check this !!!
  switch(i) {
    case 0:
      eta = -1.0;
      xi = -1.0;
    break;
    case 1:
      xi = 1.0;
      eta = -1.0;
    break;
    case 2:
      xi = 1.0;
      eta = 1.0;
    break;
    case 3:
      xi = -1.0;
      eta = 1.0;
    break;
  }
}

